from abc import ABC, abstractmethod
from time import time

class BaseModel(ABC):
    def __init__(self):
        self.index = -1
        self.last_index = False
        self.time_executed = 0  # Total elapsed time since the model started
        self.start_time = time()
        self.step_start_time = 0  # Time at the start of the current step
        self.last_value = None

    def select_variable(self, variable_name: str):
        """Select the control variable for operation."""
        try:
            for command in self.controller.commands.collection:
                if variable_name in command.variable_name:
                    self.selected_command = command
                    return
        except Exception as e:
            print(f"An error occurred while selecting the variable: {e}")

    def set_curve(self, variable_curve: list[float], duration: list[float]):
        """
        Generate a curve for the selected variable.
        :param variable_curve: A list of values defining the curve.
        :param duration: A list of durations for each value in the curve.
        """
        if not hasattr(self, "selected_command") or not self.selected_command:
            raise ValueError("No control variable selected.")
        self.curve = variable_curve
        self.duration = duration

    def run(self):
        """Run the model."""
        try:
            if hasattr(self, "curve") and hasattr(self, "duration"):
                # Check stop conditions
                if not self.last_index:
                    # Calculate elapsed time for the current step
                    elapsed_step_time = time() - self.step_start_time
                    self.time_executed = time() - self.start_time

                    duration = self.duration[self.index]
                    value = self.curve[self.index]

                    # Move to the next step if the current duration is exceeded
                    if elapsed_step_time >= duration:
                        self.step_start_time = time()  # Reset step start time
                        self.index += 1

                        # Check if we've reached the last index
                        if self.index >= len(self.curve):
                            self.last_index = True
                            return  # Stop processing further

                    # Execute command if the value changes
                    if self.last_value != value:
                        self.last_value = value
                        return self.selected_command.method(value)
                else:
                    value = 0
                    if self.last_value != value:
                        self.last_value = value
                        return self.selected_command.method(value)
            elif self.controller.node:
                self.custom_run()
                
        except Exception as e:
            print(f"An error occurred while running the model: {e}")
        
    @abstractmethod
    def custom_run(self):
        pass



from asyncua import ua

class ConfigDataAccess:
    def __init__(self, data_location: dict[str, any]):
        self.data_location = data_location
        self.devices = {}

    async def set_data_location(self):
        try:
            for device, folder in self.data_location.items():
                await self.add_device(device, folder)
        except Exception as e:
            print(f"An error occurred while setting the data location: {e}")

    async def add_device(self, device: str, folder: str):
        try:
            self.device_data = {
                'node': None,
                'folder': {
                    'node': None,
                    'variables': []
                }
            }
            self.device_data['node'] = await self.__search_node(self.client.get_objects_node(), device)
            self.device_data['folder']['node'] = await self.__search_node(self.device_data['node'], folder)
            self.devices[device] = self.device_data
            await self.__get_variables(self.device_data['folder']['node'])
        except Exception as e:
            print(f"An error occurred while adding the device: {e}")

    async def __get_variables(self, node):
        try:
            
            children_nodes = await node.get_children()
            for child_node in children_nodes:
                    node_class = await child_node.read_node_class()
                    if node_class == ua.NodeClass.Variable:
                        self.device_data['folder']['variables'].append({
                            'name': child_node.nodeid.Identifier,
                            'instance': child_node
                        })
                    else:
                        await self.__get_variables(child_node)
        except Exception as e:
            print(f"An error occurred while getting the variables: {e}")

    async def __search_node(self, node, child_name: str):
        try:
            children_nodes = await node.get_children()
            for child_node in children_nodes:
                if child_node.nodeid.Identifier == child_name:
                    return child_node
            raise Exception(f"Node {child_name} not found")
        except Exception as e:
            print(f"An error occurred while searching the node: {e}")


from connectus.tools.structure.data import DataCollection

class Configuration():
    def __init__(self):
        self.device_manager = None
        self.stop_event = None
        self.data = DataCollection()
        self.state = DataCollection()
        self.config = DataCollection()
        self.device_ids = []


from .layers import ConfigClient, ConfigDataAccess

class Configuration(ConfigClient, ConfigDataAccess): ## optimize this
    def __init__(self, node_config: dict[str, str]):
        ConfigClient.__init__(self, node_config)
        ConfigDataAccess.__init__(self, node_config['data_location'])
        self.step_time = self.node_config['step_time']
        self.type = 'communication'
        self.id = 'opc'


import asyncio

class ControllerCreator:
    def __init__(self, controller_manager, device_manager, node_creator, stop_event: asyncio.Event):
        self.controller_manager = controller_manager
        self.device_manager = device_manager
        self.node_creator = node_creator
        self.stop_event = stop_event

    def create(self, instance):
        try:
            self.controller = instance
            self._set_managers()
            if self.controller.node_params:
                self._set_node(self.controller.node_params)
        except Exception as e:
            print("An error occurred while creating controller: ", e)

    def _set_managers(self):
        self.controller.device_manager = self.device_manager
        self.controller_manager.add(self.controller)
        self.controller.stop_event = self.stop_event

    def _set_node(self, node_params: dict[str, str]):
        try:
            node = self.node_creator.create_node(node_params)
            self.controller.node = node
        except Exception as e:
            print('An error occurred while setting the node: ', e)
    


from ...base import BaseNode
from .configuration import Configuration
from connectus.tools.structure.data import VariableData
import asyncio
import socket
import datetime

class UDP(BaseNode, Configuration):
    def __init__(self, node_config: dict[str, str], stop_event: asyncio.Event):
        BaseNode.__init__(self, stop_event)
        Configuration.__init__(self, node_config)

    async def connect(self):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.bind((self.server_ip, self.server_port))
            self.sock.setblocking(False)
            self.sock.settimeout(self.timeout)
            print("UDP Server started with IP:", self.server_ip, "and port:", self.server_port)
        except Exception as e:
            print("An error occurred while starting UDP:", e)
    
    async def disconnect(self):
        self.sock.close()

    def write(self, data: list[str], node_params: dict[str, any]):
        try:
            addr = (node_params['ip_client'], node_params['port_client'])
            for msg in data:
                self.sock.sendto(msg.encode(), addr)
        except Exception as e:
            print("An error occurred while sending data with UDP protocol:", e)

    def read(self):
        try:
            msg, addr = self.sock.recvfrom(1024)
            data=VariableData(
                source=addr,
                name=msg,
                timestamp=datetime.datetime.now(datetime.timezone.utc))
            self.buffer.append(data)
        except TimeoutError:
            pass
        except asyncio.TimeoutError:
            pass
        except ConnectionResetError:
            pass
        except Exception as e:
            print("An error occurred while receiving data with UDP protocol:", e)



import pyvisa

class ConfigClient:
    def __init__(self, node_config: dict[str, str]):
        self.node_config = node_config
        self.client = pyvisa.ResourceManager(r'C:\Windows\system32\visa64.dll')
    async def create_client(self):
        try:
            await self.__set_parameters()
        except Exception as e:
            print(f"An error occurred while creating VISA client: {e}")

    async def __set_parameters(self):
        try:
            self.device = self.client.open_resource(self.resource_name, timeout=self.timeout, write_termination = self.write_termination, read_termination = self.read_termination)
        except KeyError as e:
            print(f"Error: {e} is missing in VISA server parameters.")
        except Exception as e:
            print(f"An error occurred while setting VISA parameters: {e}")


from ..layers import DataProcessing

class Configuration:
    def __init__(self, node_config: dict[str, any]):
        self.type = 'database'
        self.id = 'influxdb'
        self.url = node_config['url']
        self.token = node_config['token']
        self.org = node_config['org']
        self.bucket  = node_config['bucket']
        self.point_name = node_config['point_name']
        self.experiment_name = node_config['experiment_name']
        self.__set_layers()

    def __set_layers(self):
        self.data_processing = DataProcessing(self)


from .configuration import Configuration
from abc import ABC, abstractmethod
import asyncio

class BaseController(ABC, Configuration):
    def __init__(self, node_params: dict[str, str] = None):
        self.node_params = node_params
        self.node = None
        Configuration.__init__(self)

    async def start(self):
        print(f"Starting controller {self.id}")
        await self.initialize()
        print(f"Controller {self.id} is running")
        await self.run()
    
    @abstractmethod
    async def initialize(self):
        pass

    async def run(self):
        try:
            while self.is_running:
                data = self.acquisition.run()
                self.data_processing.process(data)
                output = self.model.run()
                self.dispatch.run(output)
                await self.check_stop()
                await asyncio.sleep(self.sample_time)
        except Exception as e:
            print(f"An error occurred while running the controller: {e}")

    async def stop(self):
        try:
            self.is_running = False
            output_data = self.data_processing.power_off()
            self.dispatch.run(output_data)
            print("Power off controller.")
        except Exception as e:
            print(f"An error occurred while stopping the controller: {e}")





class Configuration:
    def __init__(self, node_config: dict[str, any]):
        self.type = 'database'
        self.id = 'timescaledb'
        self.url = node_config['url']
        self.table_name = node_config['table_name']
        self.columns = node_config['columns']



from connectus.tools.structure.data import DataRequest
from abc import ABC, abstractmethod
import copy, time

class BaseAcquisition(ABC):
    def __init__(self):
        self.sample_rate = 1
        self.subs_time = time.time() + self.sample_rate
        self.bridge_subscription = []

    def run(self) -> list[DataRequest]:
        try:
            response_list = self.controller.device_manager.get([DataRequest(action= 'get_data', device_ids= self.controller.device_ids)])
            resquests = []
            for response in response_list:
                if response.data.collection:
                    resquests.append(DataRequest(action= 'update_data', data= response.data))
            if self.controller.node:
                resquests.extend(self.filter_messages(self.controller.node.buffer))
            if self.bridge_subscription and (time.time() - self.subs_time) >= self.sample_rate:
                resquests.extend(self.filter_messages(copy.deepcopy(self.controller.data.collection)))
                self.subs_time = time.time()
            return resquests
        except Exception as e:
            print(f"An error occurred while running controller: {e}")

    def set_bridge_subscription(self, bridge_subscription: list[str] = [], sample_rate: int = 1):
        '''
        This method is used to set the bridge subscription list
        '''
        self.bridge_subscription = bridge_subscription
        self.sample_rate = sample_rate


from .devices import DeviceManager, DeviceCreator
from .data_manager import DataManager, NodeCreator
from .controller import ControllerManager, ControllerCreator
import asyncio

class Architecture:
    def __init__(self):
        self.devices: DeviceManager = None
        self.data_manager: DataManager = None
        self.controllers: ControllerManager = None
        self.is_running = None
        self.stop_event = None
        self.general_config = {}

    async def run(self):
        print("Architecture is running")
        self.is_running = await asyncio.create_task(self.__execute())

    async def __execute(self):
        try:
            await asyncio.gather(self.data_manager.start(), self.devices.start(), self.controllers.start())
        except Exception as e:
            await self.stop()
        
    async def stop(self):
        print("Architecture is stopping")
        await self.controllers.stop()
        self.stop_event.set()
        await self.devices.stop()
        await self.data_manager.stop()
        if self.is_running:
            await self.is_running

class ArchitectureBuilder:
    def __init__(self, general_config: dict[str, any], step_factor: float = 1):
        self.architecture = Architecture()
        self.architecture.general_config = general_config
        self.architecture.stop_event = asyncio.Event()
        self.architecture.data_manager = DataManager(self.architecture.stop_event)
        self.node_creator = NodeCreator(self.architecture.data_manager, self.architecture.stop_event)
        self.architecture.devices = DeviceManager(self.architecture.stop_event, step_factor)
        self.device_creator = DeviceCreator(self.architecture.devices, self.node_creator, self.architecture.general_config)
        self.architecture.controllers = ControllerManager(self.architecture.stop_event)
        self.controller_creator = ControllerCreator(self.architecture.controllers, self.architecture.devices, self.node_creator, self.architecture.stop_event)

    def add_device(self, instance):
        self.device_creator.create(instance)

    def add_node(self, node_config: dict[str, str]):
        self.node_creator.create_node(node_config)

    def add_controller(self, instance):
        self.controller_creator.create(instance)

    def get_architecture(self):
        return self.architecture




from connectus.tools.structure.data import DataRequest
from abc import ABC, abstractmethod

class BaseDispatch(ABC):
    def __init__(self):
        pass
    
    async def send_command(self, request_list: list[DataRequest], node_params: dict[str, any]):
        if request_list:
            await self.device.node.write(request_list, node_params)
    
    async def read_command(self, request_list: list[DataRequest], node_params: dict[str, any]):
        if request_list:
            await self.device.node.read(request_list, node_params)

    


from connectus.tools.structure.data import DataRequest, DataResponse
import asyncio

class DeviceManager():
    def __init__(self, stop_event: asyncio.Event, step_factor: float = 1.0):
        self.devices = []
        self.stop_event = stop_event
        self.step_factor = step_factor
    
    def add(self, device):
        try:
            item = {}
            item['name'] = device.id
            item['instance'] = device
            self.devices.append(item)
        except Exception as e:
            print('An error occurred adding a device: ', e)

    async def start(self):
        print('Device Manager is running')
        update_bundle = []
        for device in self.devices:
            update = asyncio.create_task(self._run_device(device['instance']))
            update_bundle.append(update)
        await asyncio.gather(*update_bundle)
    
    def get(self, request_list: list[DataRequest]) -> list[DataResponse]:
        try:
            data = []
            for request in request_list:
                request_dict = request.nested_model()
                if not request_dict['device_ids']:
                    for device in self.devices:
                        data += device['instance'].get([request])
                else:
                    for device_id in request_dict['device_ids']:
                        for device in self.devices:
                            if device['name'] == device_id:
                                data += device['instance'].get([request])
                                break
            return data
        except Exception as e:
            print('An error occurred during get request: ', e)
    
    def set(self, request_list: list[DataRequest]):
        try:
            for request in request_list:
                for device_id in request.device_ids:
                    device_found = False
                    for device in self.devices:
                        if device['name'] == device_id:
                            device['instance'].set([request])
                            device_found = True
                            break
                    if not device_found:
                        print(f'Warning: Device with id {device_id} does not exist')
                    
        except Exception as e:
            print('An error occurred during set request: ', e)

    async def _run_device(self, device):
        while not self.stop_event.is_set():
            await device.run()
            await asyncio.sleep(device.step_time*self.step_factor)

    async def stop(self):
        print('Device Manager is stopping')
        for device in self.devices:
            await device['instance'].stop()
        print('All devices stopped')


from .configuration import Configuration
from connectus.tools.structure.data import DataResponse, DataRequest
from abc import ABC, abstractmethod
import asyncio

class BaseDevice(ABC, Configuration):
    def __init__(self, name: str, device_type: str, node_params: dict[str, str] = None):
        self.id = name
        self.device_type = device_type
        self.node_params = node_params
        self.node = None
        self.experiment_id = None
        Configuration.__init__(self)

    def get(self, request_list: list[DataRequest]) -> list[DataResponse]:
        try:
            response = []
            for request in request_list:
                if request.action == 'get_state':
                    response += [DataResponse(response=request.action, device_ids=[self.id], data=self.state)]
                elif request.action == 'get_config':
                    response += [DataResponse(response=request.action, device_ids=[self.id], data=self.config)]
                elif request.action == 'get_data':
                    response += [DataResponse(response=request.action, device_ids=[self.id], data=self.data)]
                else:
                    raise ValueError('Request not recognized')
            return response
        except Exception as e:
            print('An error occurred during get request: ', e)

    def set(self, request_list: list[DataRequest]):
        try:
            for request in request_list:
                requests = self.acquisition.process_request(request)
                asyncio.create_task(self.data_processing.process_data(requests))
        except Exception as e:
            print('An error occurred during set request: ', e)

    @abstractmethod
    async def custom_run(self):
        pass

    async def run(self):
        if not hasattr(self, '_init_run_done'):
            await self.init.run()
            self._init_run_done = True
        await self.custom_run()
        data = await self.acquisition.run()
        await self.data_processing.process_data(data)
        self.total_time += self.step_time

    async def stop(self):
        data = await self.acquisition.stop()
        await self.data_processing.process_data(data)



import asyncio

class DataManager:
    def __init__(self, stop_event: asyncio.Event):
        self.nodes = []
        self.stop_event = stop_event

    def add(self, node):
        try:
            item = {}
            item['name'] = node.id
            item['type'] = node.type
            item['instance'] = node
            self.nodes.append(item)
        except Exception as e:
            print('An error occurred adding a node: ', e)

    async def start(self):
        print("Data Manager is running")
        start_bundle = []
        for node in self.nodes:
            await node['instance'].connect()
            if node['type'] == 'communication':
                start_task = asyncio.create_task(node['instance'].start())
                start_bundle.append(start_task)
        if start_bundle:
            await asyncio.gather(*start_bundle)

    async def save(self, data: dict[str, dict[str, any]]): ## only database ??
        for node in self.nodes:
            if node['type'] == 'database':
                await node['instance'].write(data)

    async def stop(self):
        print("Data Manager is stopping")
        for node in self.nodes:
            await node['instance'].stop()



class Configuration:
    def __init__(self):
        self.__set_managers()

    def __set_managers(self):
        self.managers = {'device': None}

    def add_device_manager(self, device_manager: object):
        self.managers['device'] = device_manager
        self.managers['device'].add(self)

    def add_data_manager(self, data_manager: object):
        for _, device in self.devices.items():
            device.add_data_manager(data_manager)

    def _add_device(self, device: object):
        self.devices[device.id] = device


from .layers import ConfigClient

class Configuration(ConfigClient):
    def __init__(self, node_config: dict[str, str]):
        ConfigClient.__init__(self, node_config)
        self.type = 'communication'
        self.id = 'visa'
        self.timeout = node_config.get('timeout')
        self.resource_name = node_config.get('resource_name')
        self.node_id = node_config.get('node_id')
        self.write_termination = node_config.get('write_termination')
        self.read_termination = node_config.get('read_termination')


from .commands.manager import BaseCommandManager
from connectus.tools.structure.data import DataRequest, VariableData, DataCollection

class BaseDataProcessing(BaseCommandManager):
    def __init__(self):
        BaseCommandManager.__init__(self)

    async def process_data(self, request_list: list[DataRequest]):
        try:
            if request_list:
                for request in request_list:
                    if 'update_config' == request.action:
                        self._process_config(request.data)
                    elif 'update_data' == request.action:
                        await self._process_data(request.data)
                    elif 'update_state' == request.action:
                        self._process_state(request.data)
                    elif 'set_config' == request.action:
                        if self.device.device_type == 'simulated':
                            self._process_config(request.data)
                        elif self.device.device_type == 'real':
                            commands_request = self.get_commands(request.data)
                            await self.device.dispatch.send_command(commands_request, self.device.node_params)
                    elif 'error' == request.action:
                        pass
                    else:
                        raise ValueError('Data type not recognized during processing data')
        except Exception as e:
            print('An error occurred during processing data: ', e)

    def _process_config(self, data_collection: DataCollection):
        ''' Check if the configuration is correct and update the device configuration'''
        try:
            if data_collection:
                for data in data_collection.collection:
                    self.device.config.update(data)
        except Exception as e:
            print('An error occurred while processing the configuration: ', e)

    def _process_state(self, data_collection: DataCollection):
        ''' Check if the configuration is correct and update the device configuration'''
        try:
            if data_collection:
                for data in data_collection.collection:
                    self.device.state.update(data)
        except Exception as e:
            print('An error occurred while processing the configuration: ', e)
    
    async def _process_data(self, data_collection: DataCollection):
        ''' Update the device data with the new values and save them in the database'''
        try:
            if data_collection:
                for data in data_collection.collection:
                    self.device.data.update(data)
                    await self.device.managers['node'].save(data) ## esto es eficiente ?? check if it is better to use asyncio.create_task or await
        except Exception as e:
            print('An error occurred while processing the data update: ', e)
    



from ...base import BaseNode
from .configuration import Configuration
from connectus.tools.structure.data import DataRequest, VariableData
import asyncio
import datetime

class VISA(BaseNode, Configuration):
    def __init__(self, node_config: dict[str, str], stop_event: asyncio.Event):
        BaseNode.__init__(self, stop_event)
        Configuration.__init__(self, node_config)

    async def connect(self):
        try:
            await self.create_client()
        except Exception as e:
            print("An error occurred while starting VISA:", e)
    
    async def disconnect(self):
        try:
            await asyncio.sleep(1)  # Wait for 1 second before closing the connection
            self.device.close()
            self.client.close()
        except Exception as e:
            print("An error occurred while disconnecting VISA:", e)

    async def write(self, request_list: list[DataRequest], node_params: dict[str, any]= None):
        try:
            if request_list:
                for request in request_list:
                    for variable in request.data.collection:
                        self.device.write(variable.value)
        except Exception as e:
            print("An error occurred while sending data with VISA protocol:", e)

    async def read(self, request_list: list[DataRequest] =[], node_params: dict[str, any] = None):
        try:
            if request_list:
                for request in request_list:
                    for variable in request.data.collection:
                        value = await self._async_query(variable)
                        self.buffer.append(value)

        except Exception as e:
            print(f"An error occurred while reading the data: {e}")

    async def _async_query(self, variable):
        """Helper method to execute query_ascii_values asynchronously."""
        try:
            value = await asyncio.to_thread(self.device.query_ascii_values, variable.value)
            return VariableData(
                source="visa",
                name=variable.name,
                value=value[0] if value else None,
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                value_type=type(value[0]).__name__ if value else "None",
            )
        except Exception as e:
            return None  # Return None to avoid breaking `gather()`


from connectus.tools.structure.data import DataRequest, VariableData, DataCollection
from abc import ABC, abstractmethod
from datetime import datetime, timezone

class BaseAcquisition(ABC):
    def __init__(self):
        pass

    async def run(self) -> list[DataRequest]:
        if self.device.device_type == 'simulated':
            data = self.device.model.run()
            return data
        elif self.device.device_type == 'real':
            data = self.filter_messages(self.device.node.buffer)
            return data
                
    def process_request(self, request: DataRequest) -> list[DataRequest]:
        ''' convert data to the right format (power to voltage, etc.) '''
        try:
            requests = []
            if 'set_config' == request.action: ## we need to check if the request is valid (is a configurable value?)
                requests.append(request)
            else:
                raise ValueError('Invalid request type')
            return requests
        except Exception as e:
            print('An error occurred during processing request: ', e)
    
    async def stop(self) -> list[DataRequest]: ## We send the data to the database as NULL
        ''' stop the acquisition '''
        data = []
        for source, variables in self.device.data.nested_model().items():
            for name in variables:
                data.append(VariableData(
                    name= name, 
                    timestamp= datetime.now(timezone.utc), 
                    source= source, 
                    experiment_id= variables[name].get('experiment_id'), 
                    unit= variables[name].get('unit'), 
                    value_type= variables[name].get('value_type'), 
                    additional_info= variables[name].get('additional_info')
                ))

        return [DataRequest(action= 'update_data', device_ids= [self.device.id], data= DataCollection(collection=data))]
        


from connectus.tools.structure.data import DataRequest, DataCollection
from abc import ABC, abstractmethod
import json

class BaseDispatch(ABC):
    def __init__(self):
        pass

    def run(self, request_list: list[DataRequest]):
        if request_list:
            self.controller.device_manager.set(request_list)

    def send_bridge_data(self, data_collection: DataCollection):
        if data_collection.collection:
            for variable_data in data_collection.collection:
                data = {variable_data.name: variable_data.value}
                json_data = json.dumps(data)
                self.controller.node.write([json_data], self.controller.node_params)


from connectus.tools.structure.data import DataCollection

class Configuration:
    def __init__(self):
        self.total_time = 0
        self._set_managers()
        self._set_variables()

    def _set_managers(self):
        self.managers = {'device': None, 'node': None}

    def add_device_manager(self, device_manager):
        self.managers['device'] = device_manager
        self.managers['device'].add(self)
    
    def add_data_manager(self, data_manager):
        self.managers['node'] = data_manager

    def _set_variables(self):
        self.data = DataCollection() # Received data from the real device (saved in the database)
        self.state = DataCollection() # State of the device (read only)
        self.config = DataCollection() # Configuration of the device (read/write)


from connectus.tools.structure.data import DataRequest, VariableData
from abc import ABC, abstractmethod

class BaseCommandManager(ABC):
    def __init__(self):
        pass

    def _get_state_value(self, name):
        """Retrieve the current state value for a specific field name."""
        for field in self.device.state:
            if field["name"] == name:
                return field["value"]
        return None  # Return None if name is not found

    def _is_value_within_limits(self, name, value):
        """Check if a value is within its allowed limits."""
        for limit in self.limits:
            if limit["name"] == name:
                return limit["min_value"] <= value <= limit["max_value"]
        return True  # No limits defined; assume valid

    def _is_command_needed(self, name, new_value):
        """Check if a command is needed based on the current state."""
        current_value = self._get_state_value(name)
        return current_value != new_value

    def get_commands(self, data: list[VariableData]) -> list[DataRequest]: # we must check which commands we must send to the device to set the new value (black box)
        
        # command = next(
        #     (cmd for cmd in self.device.config if cmd["name"] == name),
        #     None,
        # )
        # if not command:
        #     raise ValueError(f"Unknown command field: {name}")

        # if not self._is_value_within_limits(name, new_value):
        #     raise ValueError(f"Value {new_value} is out of allowed limits for '{name}'.")

        # if not self._is_command_needed(name, new_value):
        #     print(f"Command '{name}' with value '{new_value}' is not needed.")
        #     return
        commands_request = [
            DataRequest(
                action="set_config",
                device_ids=[self.device.id],
                data=data)
        ]
        return commands_request


from abc import ABC, abstractmethod
import asyncio

class BaseNode(ABC):
    def __init__(self, stop_event: asyncio.Event):
        self.stop_event = stop_event
        self.buffer = []

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    async def start(self, sample_time: int = 1): ## indicate what is start
        try:
            while not self.stop_event.is_set():
                self.read() 
                await asyncio.sleep(sample_time)
            await self.stop()
        except Exception as e:
            print('An error occurred during running a node: ', e)

    async def stop(self):
        try:
            self.stop_event.set()
            await self.disconnect()
        except Exception as e:
            print('An error occurred stopping a node: ', e)

    @abstractmethod
    async def read(self):
        pass

    @abstractmethod
    def write(self):
        pass
    



class Configuration:
    def __init__(self, node_config: dict[str, str]):
        self.type = 'communication'
        self.id = 'udp'
        self.timeout = node_config['timeout']
        self.server_ip = node_config['ip']
        self.server_port = node_config['port']  # default server port


from connectus.tools.structure.data import DataCollection, DataRequest
from abc import ABC, abstractmethod

class BaseDataProcessing(ABC):
    def __init__(self):
        pass
    
    def process(self, request_list: list[DataRequest]):
        try:
            if request_list:
                for request in request_list:
                    if 'update_data' == request.action:
                        self._process_data(request.data)
                    elif 'set_config' == request.action:
                        self._process_config(request)
                    elif 'bridge_data' == request.action:
                        self._process_bridge_data(request.data)
                    elif 'error' == request.data:
                        pass
                    else:
                        raise ValueError('Data type not recognized during processing data')
        except Exception as e:
            print('An error occurred during processing data: ', e)

    def _process_data(self, data_collection: DataCollection):
        ''' Update the device data with the new values and save them in the database'''
        try:
            if data_collection.collection:
                self.controller.data.update(data_collection)
        except Exception as e:
            print('An error occurred while processing the data update: ', e)

    def _process_config(self, requests: DataRequest):
        try:
            self.controller.dispatch.run([requests])
        except Exception as e:
            print('An error occurred while processing the configuration: ', e)

    def _process_bridge_data(self, data_collection: DataCollection):
        try:
            self.controller.dispatch.send_bridge_data(data_collection)
        except Exception as e:
            print('An error occurred while processing bridge data: ', e)


from abc import ABC, abstractmethod

class BaseInit(ABC):
    """
    Base class for initialization layers.
    """

    def __init__(self) -> None:
        """
        Initialize the base class.
        """
        super().__init__()

    @abstractmethod
    async def run(self) -> None:
        """
        Abstract method to run the initialization process.
        Must be implemented by subclasses.
        """
        pass


from connectus.tools.structure.data import DataRequest
from abc import ABC, abstractmethod

class BaseModel(ABC):
    def __init__(self):
        pass
    
    def run(self) -> list[DataRequest]:
        try:
            points = []
            points.append(DataRequest(action= 'update_data', data= self.get_data()))
            return points
        except Exception as e:
            print('An error occurred during model run: ', e)


from ...base import BaseNode
from .configuration import Configuration
import asyncio
import serial

class UART(BaseNode, Configuration):
    def __init__(self, node_config: dict[str, str], stop_event: asyncio.Event):
        BaseNode.__init__(self, stop_event)
        Configuration.__init__(self, node_config)

    async def connect(self):
        try:
            self.ser = serial.Serial(self.port, self.baudrate, self.bytesize, self.parity, self.stopbits, self.timeout)
            print("UART Server started with port: ", self.port)
        except Exception as e:
            print("An error ocurred while starting UART: ", e)
    
    async def disconnect(self):
        self.ser.close()

    def write(self, data: list[str], node_params: dict[str, any]):
        try:
            for msg in data:
                self.ser.write(msg)
        except Exception as e:
            print("An error ocurred while sending data with UART protocol: ", e)

    def read(self):
        try:
            line = self.ser.readline()
            return line
        except TimeoutError:
            pass
        except Exception as e:
            print("An error occurred while receiving data:", e)


import asyncua as opc

class ConfigClient:
    def __init__(self, node_config: dict[str, str]):
        self.node_config = node_config
        self.client = opc.Client(url=node_config['url'], timeout=int(self.node_config['timeout']))  # Initialize with the URL
        self.certificate_path = None
        self.private_key_path = None

    async def create_client(self):
        try:
            await self.__set_parameters()
            await self.client.connect()
        except Exception as e:
            print(f"An error occurred while creating OPC client: {e}")

    async def __set_parameters(self):
        try:
            self.client.application_uri = self.node_config['app_uri']
            self.client.set_user(self.node_config.get('user', ""))
            self.client.set_password(self.node_config.get('password', ""))
            await self.client.set_security_string(
                f"{self.node_config['policy']},"
                f"{self.node_config['mode']},"
                f"{self.node_config.get('certificate', '')},"
                f"{self.node_config.get('private_key', '')}"
            )
        except KeyError as e:
            print(f"Error: {e} is missing in OPC server parameters.")
        except Exception as e:
            print(f"An error occurred while setting OPC parameters: {e}")


from ..type.custom import UART, UDP, InfluxDB, OPC, TimeScaleDB, VISA
import asyncio

class NodeCreator:
    def __init__(self, data_manager, stop_event: asyncio.Event):
        self.stop_event = stop_event
        self.data_manager = data_manager
        self.id = 0

    def create_node(self, node_config: dict[str, str]):
        node_config['id'] = self.id
        name = node_config['name']
        if name == 'uart':
            node = UART(node_config, self.stop_event)
        elif name == 'udp':
            node = UDP(node_config, self.stop_event)
        # elif name == 'influxdb':
        #     node = InfluxDB(node_config, self.stop_event)
        elif name == 'opc':
            node = OPC(node_config, self.stop_event)
        elif name == 'timescaledb':
            node = TimeScaleDB(node_config, self.stop_event)
        elif name == 'visa':
            node = VISA(node_config, self.stop_event)
        else:
            raise ValueError(f"Node type {name} not supported")
        self.data_manager.add(node)
        self.id += 1
        return node


from .init.init import BaseInit
from .acquisition.acquisition import BaseAcquisition
from .data_processing.data_processing import BaseDataProcessing
from .model.model import BaseModel
from .dispatch.dispatch import BaseDispatch


from connectus.tools.structure.data import VariableData
from ...base import BaseNode
from .configuration import Configuration
import asyncpg
import asyncio

class TimeScaleDB(BaseNode, Configuration):
    def __init__(self, node_config: dict[str, any], stop_event: asyncio.Event):
        BaseNode.__init__(self, stop_event)
        Configuration.__init__(self, node_config)
        self.pool = None  # Connection pool

    async def connect(self):
        """Initialize the connection pool."""
        try:
            self.pool = await asyncpg.create_pool(self.url)  # Use connection pool
            async with self.pool.acquire() as conn:
                query = self._create_table(self.table_name, self.columns)
                await conn.execute(query)
        except Exception as e:
            print(f"An error occurred while connecting to the database: {e}")

    async def write(self, data: VariableData, table_name: str= None):
        """
        Insert data into the specified table with validation against the table's schema.

        Args:
            table_name (str): The name of the table to insert data into.
            data (list[dict[str, any]]): A list of dictionaries, where each dictionary represents a row to insert.
                                         Keys in the dictionary should match column names in the table.

        Raises:
            ValueError: If column names in the data do not match the table's schema.
        """
        table_name = table_name or self.table_name

        if not data:
            print("No data to write.")
            return
        
        data = data.plain_model()

        async with self.pool.acquire() as conn:
            try:
                # Fetch the table schema
                query = f"""
                    SELECT column_name
                    FROM information_schema.columns
                    WHERE table_name = $1;
                """
                result = await conn.fetch(query, table_name.lower())
                valid_columns = {row['column_name'] for row in result}

                # Get column names from the first dictionary in the data
                data_columns = set(data.keys())

                # Check for invalid columns
                invalid_columns = data_columns - valid_columns
                if invalid_columns:
                    await self._alter_table(conn, table_name, invalid_columns)

                # Prepare columns and placeholders
                column_names = ", ".join(data_columns)
                placeholders = ", ".join(f"${i+1}" for i in range(len(data_columns)))

                # Construct the query
                query = f"INSERT INTO {table_name} ({column_names}) VALUES ({placeholders})"

                # Prepare the data for insertion
                values = [
                    tuple(str(data[col]) if col != 'timestamp' else data[col] for col in data_columns)
                ]

                # Execute the query for each row of data
                for value_set in values:
                    await conn.execute(query, *value_set)

            except ValueError as ve:
                print(f"Validation error: {ve}")
            except Exception as e:
                print(f"An error occurred while inserting data into {table_name}: {e}")

    async def disconnect(self):
        """Close the connection pool properly."""
        if self.pool:
            await self.pool.close()

    def read(self):
        pass

    def _create_table(self, name: str, columns: list[tuple[str, str]]):
        """
        Create a table with the given name and columns.
        
        Args:
            name (str): The name of the table to create.
            columns (list[tuple[str, str]]): A list of tuples, where each tuple contains a column name and its SQL type.
        
        Example:
            columns = [
                ('time', 'TIMESTAMPTZ NOT NULL'),
                ('exp_id', 'TEXT NOT NULL'),
                ('name', 'TEXT NOT NULL'),
                ('value', 'TEXT'),
                ('type', 'TEXT NOT NULL')
            ]
        """
        try:
            columns_def = ", ".join([f"{col_name} {col_type}" for col_name, col_type in columns])
            query = f"CREATE TABLE IF NOT EXISTS {name} ({columns_def});"
            return query
        except Exception as e:
            print(f"An error occurred while creating the table {name}: {e}")
            
    async def _alter_table(self, conn, table_name: str, new_columns: set[str]):
        """
        Alter the table to add new columns if they do not exist.
        """
        try:
            query = """
                SELECT column_name
                FROM information_schema.columns
                WHERE table_name = $1;
            """
            result = await conn.fetch(query, table_name.lower())
            existing_columns = {row['column_name'] for row in result}

            columns_to_add = new_columns - existing_columns
            if not columns_to_add:
                return

            for column in columns_to_add:
                alter_query = f"ALTER TABLE {table_name} ADD COLUMN {column} TEXT;"
                await conn.execute(alter_query)

        except Exception as e:
            print(f"An error occurred while altering the table {table_name}: {e}")


from ...base import BaseNode
from .configuration import Configuration
from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
import asyncio

class InfluxDB(BaseNode, Configuration):
    def __init__(self, node_config: dict[str, any], stop_event: asyncio.Event):
        BaseNode.__init__(self, stop_event)
        Configuration.__init__(self, node_config)

    async def connect(self):
        try:
            self.client = InfluxDBClientAsync(url=self.url, token=self.token, org=self.org)
        except Exception as e:
            print(f"An error occurred while connecting to the database: {e}")

    async def write(self, data: list[dict[str, any]]):
        try:
            if data:
                points = self.data_processing.prepare_data(data)
                write_api = self.client.write_api()
                await write_api.write(bucket=self.bucket, record=points)
        except Exception as e:
            print(f"An error occurred while inserting the data: {e}")

    async def disconnect(self):
        await self.client.close()

    def read(self):
        pass


import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

VERSION = '0.0.18' #Muy importante, deberéis ir cambiando la versión de vuestra librería según incluyáis nuevas funcionalidades
PACKAGE_NAME = 'connectus' #Debe coincidir con el nombre de la carpeta 
AUTHOR = 'Javier A. Quintana Hernandez' #Modificar con vuestros datos
AUTHOR_EMAIL = 'javierquinttana@gmail.com' #Modificar con vuestros datos
URL = 'https://github.com/javierquinttana' #Modificar con vuestros datos

LICENSE = 'MIT' #Tipo de licencia
DESCRIPTION = 'Librería para la gestión de dispositivos físicos y virtuales.' #Descripción corta
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding='utf-8') #Referencia al documento README con una descripción más elaborada
LONG_DESC_TYPE = "text/markdown"


#Paquetes necesarios para que funcione la libreía. Se instalarán a la vez si no lo tuvieras ya instalado
INSTALL_REQUIRES = [
    'pyserial',
    'asyncua',
    'cryptography',
    'aiohttp',
    'aiocsv',
    'pandas',
    'numpy',
    'matplotlib',
    'asyncpg',
    'pyvisa',
    'psutil',
    'zeroconf',
    'paho-mqtt'
      ]

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    install_requires=INSTALL_REQUIRES,
    license=LICENSE,
    packages=find_packages(),
    include_package_data=True
)


from dataclasses import dataclass, field
from typing import Optional, Dict, Any, List, Union
from datetime import datetime

@dataclass
class CommandData:
    variable_name: str = field(metadata={"description": "The name of the variable"})
    method: callable = field(metadata={"description": "The method to be executed"})

    def nested_model(self) -> Dict[str, Dict[str, Any]]:
        """Converts the dataclass instance to a nested dictionary."""
        data = {self.variable_name: {"method": self.method}}
        return data

@dataclass
class CommandCollection:
    collection: List[CommandData] = field(default_factory=list)

    def _deep_update(self, source, updates):
        for key, value in updates.items():
            if isinstance(value, dict) and key in source:
                self._deep_update(source[key], value)
            else:
                source[key] = value

    def _update_if_exists(self, data: CommandData):
        for existing_data in self.collection:
            if existing_data.variable_name == data.variable_name:
                self.collection.remove(existing_data)
                break
        self.collection.append(data)

    def update(self, data: Union[CommandData, 'CommandCollection']):
        """Adds a CommandData instance to the collection."""
        if isinstance(data, CommandData):
            self._update_if_exists(data)
        elif isinstance(data, CommandCollection):
            for new_data in data.collection:
                self._update_if_exists(new_data)

    def nested_model(self) -> Dict[str, Dict[str, Any]]:
        output = {}
        for data in self.collection:
            self._deep_update(output, data.nested_model())
        return output

@dataclass
class VariableData:
    source: str = field(metadata={"description": "Source identifier"})
    name: str = field(metadata={"description": "The name of the variable"})
    value: Optional[float] = field(default=None, metadata={"description": "The value of the variable"})
    timestamp: datetime = field(
        default_factory=datetime.now,
        metadata={"description": "Timestamp when the data was recorded"}
    )
    value_type: Optional[Any] = field(default=None, metadata={"description": "Type of the data value"})
    experiment_id: Optional[str] = field(default=None, metadata={"description": "Experiment identifier"})
    unit: Optional[str] = field(default=None, metadata={"description": "Unit of the data value, if applicable"})
    additional_info: Optional[Dict[str, Any]] = field(
        default_factory=dict, metadata={"description": "Additional optional information"}
    )

    def __post_init__(self):
        if self.value_type is None and self.value is not None:
            self.value_type = type(self.value).__name__

    def nested_model(self) -> Dict[str, Any]:
        """Converts the dataclass instance to a nested dictionary."""
        data = {
            self.source: {
                self.name: {
                    "timestamp": self.timestamp,
                    "additional_info": self.additional_info,
                }
            }
        }
        if self.value is not None:
            data[self.source][self.name]["value"] = self.value
        if self.experiment_id is not None:
            data[self.source][self.name]["experiment_id"] = self.experiment_id
        if self.unit is not None:
            data[self.source][self.name]["unit"] = self.unit
        if self.value_type is not None:
            data[self.source][self.name]["value_type"] = self.value_type
        return data
    
    def plain_model(self) -> Dict[str, Any]:
        """Converts the dataclass instance to a nested dictionary."""
        data = {
            "source": self.source,
            "name": self.name,
            "timestamp": self.timestamp,
            "additional_info": self.additional_info,
        }
        if self.value is not None:
            data["value"] = self.value
        if self.experiment_id is not None:
            data["experiment_id"] = self.experiment_id
        if self.unit is not None:
            data["unit"] = self.unit
        if self.value_type is not None:
            data["value_type"] = self.value_type
        return data

@dataclass
class DataCollection:
    collection: List[VariableData] = field(default_factory=list)

    def _deep_update(self, source, updates):
        for key, value in updates.items():
            if isinstance(value, dict) and key in source:
                self._deep_update(source[key], value)
            else:
                source[key] = value

    def _update_if_exists(self, data: VariableData):
        for existing_data in self.collection:
            if existing_data.source == data.source and existing_data.name == data.name:
                self.collection.remove(existing_data)
                break
        self.collection.append(data)

    def update(self, data: Union[VariableData, 'DataCollection']):
        """Adds a VariableData instance to the collection."""
        if isinstance(data, VariableData):
            self._update_if_exists(data)
        elif isinstance(data, DataCollection):
            for new_data in data.collection:
                self._update_if_exists(new_data)

    def nested_model(self) -> Dict[str, Dict[str, Dict[str, Any]]]:
        output = {}
        for data in self.collection:
            self._deep_update(output, data.nested_model())
        return output
    
    def plain_model(self) -> List[Dict[str, Any]]:
        output = []
        for data in self.collection:
            output.append(data.plain_model())
        return output

@dataclass
class BaseExchange:
    device_ids: List[str] = field(default_factory=list)
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())

@dataclass
class DataRequest(BaseExchange):
    action: str = field(default_factory=str, metadata={"description": "The action to be performed"})
    data: Optional [DataCollection] = field(default=None)

    def nested_model(self) -> Dict[str, Any]:
        """Converts the dataclass instance to a dictionary."""
        data = {
            "action": self.action,
            "device_ids": self.device_ids,
            "timestamp": self.timestamp,
            "data": self.data,
        }
        return data

@dataclass
class DataResponse(BaseExchange):
    response: str = field(default_factory=str, metadata={"description": "The response to the request"})
    data: Optional[DataCollection] = field(default=None)
    error_message: Optional[str] = field(default=None)

    def nested_model(self) -> Dict[str, Any]:
        """Converts the dataclass instance to a dictionary."""
        data = {
            "response": self.response,
            "device_ids": self.device_ids,
            "timestamp": self.timestamp,
            "data": self.data,
            "error_message": self.error_message,
        }
        return data



import asyncio

class ControllerManager:
    def __init__(self, stop_event: asyncio.Event):
        self.controllers = []
        self.stop_event = stop_event

    def add(self, controller):
        try:
            item = {}
            item['id'] = controller.id
            item['instance'] = controller
            self.controllers.append(item)
        except Exception as e:
            print('An error occurred adding a controller: ', e)

    async def start(self): ## all controllers are started in parallel
        start_bundle = []
        for controller in self.controllers:
            start_task = asyncio.create_task(controller['instance'].start())
            start_bundle.append(start_task)
        if start_bundle:
            await asyncio.gather(*start_bundle)

    async def stop(self):
        for controller in self.controllers:
            await controller['instance'].stop()



class Configuration:
    def __init__(self, node_config: dict[str, str]):
        self.type = 'communication'
        self.id = 'uart'
        self.port = node_config['port']
        self.baudrate = node_config['baudrate']
        self.bytesize = node_config['bytesize']
        self.parity = node_config['parity']
        self.stopbits = node_config['stopbits']
        self.timeout = node_config['timeout'] ## We must use it ?
        self.addr = self.port


from connectus.tools.structure.data import DataRequest
from connectus.data_manager.node.type.base import BaseNode
from .configuration import Configuration
from .configuration.tools.subscription import SubscriptionHandler
from asyncua import ua
import asyncio

class OPC(BaseNode, Configuration):
    def __init__(self, node_config: dict[str, str], stop_event: asyncio.Event):
        BaseNode.__init__(self, stop_event)
        Configuration.__init__(self, node_config)
    
    async def connect(self):
        try:
            await self.create_client()
            await self.set_data_location()
            self.subscription = await self.create_subscription(1000)
        except Exception as e:
            raise ConnectionError(f"An error occurred while connecting to the OPC server: {e}")
        
    async def disconnect(self):
        try:
            if await self.client.check_connection():
                await self.client.disconnect()
        except Exception as e:
            print(f"An error occurred while disconnecting from the OPC server: {e}")

    async def create_subscription(self, period: int):
        try:
            handler = SubscriptionHandler(self.buffer)
            self.subscription = await self.client.create_subscription(period, handler)
            nodes = []
            for _, data in self.devices.items():
                for variable in data['folder']['variables']:
                    nodes.append(variable['instance'])
            await self.subscription.subscribe_data_change(nodes)
        except Exception as e:
            print(f"An error occurred while creating the subscription: {e}")

    def read(self):
        try:
            pass
        except Exception as e:
            print(f"An error occurred while reading the data: {e}")

    async def write(self, request_list: list[DataRequest], node_params: dict[str, any]= None): ## include check if variable exists in the server/device
        try:
            if request_list:
                for request in request_list:
                    for variable in request.data.collection:
                        for device in self.devices:
                            for opc_variable in self.devices[device]['folder']['variables']:
                                if variable.name in opc_variable['name']:
                                    data_type_attribute = await opc_variable['instance'].read_attribute(ua.AttributeIds.DataType)
                                    data_type = ua.VariantType(data_type_attribute.Value.Value.Identifier)
                                    data_value = ua.DataValue(ua.Variant(variable.value, data_type))
                                    await opc_variable['instance'].set_value(data_value)
        except Exception as e:
            print(f"An error occurred while writing the data: {e}")




from connectus.tools.structure.data import VariableData
from datetime import timezone

class SubscriptionHandler(object):
    """
    Subscription Handler. To receive events from server for a subscription
    This class is just a sample class. Whatever class having these methods can be used
    """
    def __init__(self, changes_buffer):    
        self.buffer = changes_buffer  # To store updates

    def data_change(self, handle, node, value, attr):
        """
        Deprecated, use datachange_notification
        """
        pass

    def datachange_notification(self, node, value, data):
        """
        called for every datachange notification from server
        """
        item = VariableData(
            source= "opcua",
            name= node.nodeid.Identifier,
            value= value,
            timestamp= data.monitored_item.Value.SourceTimestamp.replace(tzinfo=timezone.utc),
            value_type= type(value).__name__,
        )

        self.buffer.append(item)

    def event_notification(self, event):
        """
        called for every event notification from server
        """
        pass

    def status_change_notification(self, status):
        """
        called for every status change notification from server
        """
        pass



from abc import ABC, abstractmethod
from .configuration import Configuration

class BasePack(ABC, Configuration): ## don't have node implementation
    def __init__(self, name: str, device_type: str, node_params: dict[str, str] = None):
        self.id = name
        self.device_type = device_type
        self.node_params = node_params
        self.node = None
        self.devices = {}
        Configuration.__init__(self)

    @abstractmethod
    def run(self):
        pass

    def get(self, request: list[dict[str, str]]) -> list[dict[str, any]]:
        ''' request = [{'action': 'get_data'}]
            response = [{'response': 'get_data', 'device_id': device_id, 'data': [{...}, {...}]}] '''
        try:
            output = []
            for _, device in self.devices.items():
                    output += device.get(request)
            return output
        except Exception as e:
            print('An error occurred during get request: ', e)
    
    def set(self, request: list[dict[str, any]]):
        ''' request = [{'action': 'set_config', 'data': [{'name': name, 'value': value}]}] '''
        try:
            for item in request:
                for point in item['data']:
                    for _, device in self.devices.items():
                        if device.id == point['device_id']:
                            device.set([point])
        except Exception as e:
            print('An error occurred during set request: ', e)


from influxdb_client import Point
from datetime import datetime

class DataProcessing:
    def __init__(self, node):
        self.node = node
        self.initial_time = str(datetime.now())

    def prepare_data(self, input_data: list[dict[str, any]]):
        ''' Convert the data to the correct format for the database'''
        output_data = []
        for measurement in input_data:
            point = Point(self.node.point_name)
            name = None
            value = None
            for name, value in measurement.items():
                if name == 'name':
                    name = value
                elif name == 'value':
                    value = value
                else:
                    point.tag(name, value)
            if name != None and value != None:
                point.field(name, value)
            else:
                raise ValueError('Field name or field value not found in the data')
            point.tag('experiment', self.node.experiment_name + ' ' + self.initial_time)
            output_data.append(point)
        return output_data



class DeviceCreator():
    def __init__(self, device_manager, node_creator, general_config: dict[str, any]):
        self.device_manager = device_manager
        self.node_creator = node_creator
        self.general_config = general_config

    def create(self, instance):
        try:
            self.device = instance
            self._set_device_manager()
            self._set_data_manager()
            self._set_general_config()
            if self.device.node_params:
                self._set_node(self.device.node_params)
        except Exception as e:
            print("An error occurred while creating device: ", e)
            
    def _set_device_manager(self):
        try:
            self.device.add_device_manager(self.device_manager)
        except Exception as e:
            print('An error occurred while setting the device manager: ', e)

    def _set_data_manager(self):
        try:
            self.device.add_data_manager(self.node_creator.data_manager)
        except Exception as e:
            print('An error occurred while setting the data manager: ', e)
    
    def _set_node(self, node_params: dict[str, str]):
        try:
            node = self.node_creator.create_node(node_params)
            self.device.node = node
        except Exception as e:
            print('An error occurred while setting the node: ', e)
    
    def _set_general_config(self):
        self.device.experiment_id = self.general_config['experiment_id']